% This code has been taken from website www.mathworks.com, and was
% modified by myself, by changing something or by adding some features.


% Info on the author of the code taken from www.mathworks.com:

% Author: Joseph Kirk
% Email: jdkirk630 at gmail dot com
% Release: 1.3
% Release Date: 5/18/07


% License present on the code taken from www.mathworks.com:

% Copyright (c) 2006, Joseph Kirk
% All rights reserved.
% 
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are
% met:
% 
%     * Redistributions of source code must retain the above copyright
%       notice, this list of conditions and the following disclaimer.
%     * Redistributions in binary form must reproduce the above copyright
%       notice, this list of conditions and the following disclaimer in
%       the documentation and/or other materials provided with the distribution
% 
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
% IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
% ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
% LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
% SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
% INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
% CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
% ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
% POSSIBILITY OF SUCH DAMAGE.


function [dist,path] = dijkstra_algorithm(nodes,edges,start_id,finish_id)


node_ids = nodes(:,1);
[num_map_pts,cols] = size(nodes);
table = sparse(num_map_pts,2);
shortest_distance = Inf(num_map_pts,1);
settled = zeros(num_map_pts,1);
path = num2cell(NaN(num_map_pts,1));
col = 2;
pidx = find(start_id == node_ids);
shortest_distance(pidx) = 0;
table(pidx,col) = 0;
settled(pidx) = 1;
path(pidx) = {start_id};
if (nargin < 4) % compute shortest path for all nodes
    while_cmd = 'sum(~settled) > 0';
else % terminate algorithm early
    while_cmd = 'settled(zz) == 0';
    zz = find(finish_id == node_ids);
end
while eval(while_cmd)
    % update the table
    table(:,col-1) = table(:,col);
    table(pidx,col) = 0;
    % find neighboring nodes in the segments list
    neighbor_ids = [edges(node_ids(pidx) == edges(:,2),3);
        edges(node_ids(pidx) == edges(:,3),2)];
    % calculate the distances to the neighboring nodes and keep track of the paths
    for k = 1:length(neighbor_ids)
        cidx = find(neighbor_ids(k) == node_ids);
        if ~settled(cidx)
            d = sqrt(sum((nodes(pidx,2:cols) - nodes(cidx,2:cols)).^2));
            if (table(cidx,col-1) == 0) || ...
                    (table(cidx,col-1) > (table(pidx,col-1) + d))
                table(cidx,col) = table(pidx,col-1) + d;
                tmp_path = path(pidx);
                path(cidx) = {[tmp_path{1} neighbor_ids(k)]};
            else
                table(cidx,col) = table(cidx,col-1);
            end
        end
    end
    % find the minimum non-zero value in the table and save it
    nidx = find(table(:,col));
    ndx = find(table(nidx,col) == min(table(nidx,col)));
    if isempty(ndx)
        break
    else
        pidx = nidx(ndx(1));
        shortest_distance(pidx) = table(pidx,col);
        settled(pidx) = 1;
    end
end
if (nargin < 4) % return the distance and path arrays for all of the nodes
    dist = shortest_distance';
    path = path';
else % return the distance and path for the ending node
    dist = shortest_distance(zz);
    path = path(zz);
    path = path{1};
end


end
